open Big_int
open Utils
open Sha256
open Secp256k1

exception Invalid

let aimprivkey k =
  let cpriv = base58 k in
  let sh1 = sha256 cpriv in
  "AK" ^ cpriv ^ (String.sub sh1 0 8)

let privkey_from_aimprivkey ak =  
  let akl = String.length ak in
  if not (akl > 12) then raise (Failure "invalid AIM private key");
  if not (ak.[0] = 'A') then raise (Failure "invalid AIM private key");
  if not (ak.[1] = 'K') then raise (Failure "invalid AIM private key");
  try
    frombase58 (String.sub ak 2 (akl - 10))
  with (Failure(_)) -> (*** try an older hex format ***)
    big_int_of_hex (String.sub ak 2 (akl - 10))

let valid_aimprivkey ak =
  let akl = String.length ak in
  if akl > 12 && ak.[0] = 'A' && ak.[1] = 'K' then
    begin
      try
	let k = frombase58 (String.sub ak 2 (akl - 10)) in
	ak = aimprivkey k
      with Failure(_) -> false
    end
  else
    false

(* Computation of nym from public key (for AIM, include full pubkey in nym, plus checksum) *)

(* (x,y) : public key (as a nonzero point on the elliptic curve) *)
let aimnym (x,y) =
  let cpubx = base58 x in
  let cpuby = base58 y in
  let sh1 = sha256 (cpubx ^ ":" ^ cpuby) in
  (cpubx,cpuby,base58 (big_int_of_hex (String.sub sh1 0 8)))

let pubkey_from_aimnym nym =
  let (ax,ay,ac) = nym in
  let x = frombase58 ax in
  let y = frombase58 ay in
  (x,y)

let valid_aimnym_p nym =
  let (ax,ay,ac) = nym in
  let x = frombase58 ax in
  let y = frombase58 ay in
  nym = aimnym(x,y)

let inv2 x q =
  let (u,v) = eea q (mod_big_int x q) in
  mod_big_int v q

let aimsgn (r,s) =
  "AS" ^ (hex_of_big_int r 32) ^ (hex_of_big_int s 32)

let sgn_from_aimsgn asg =
  if not (String.length asg = 130) then raise (Failure "invalid AIM signature");
  if not (asg.[0] = 'A') then raise (Failure "invalid AIM signature");
  if not (asg.[1] = 'S') then raise (Failure "invalid AIM signature");
  let r = big_int_of_hex (String.sub asg 2 64) in
  let s = big_int_of_hex (String.sub asg 66 64) in
  (r,s)

let make_sgn privk randk msgnum =
  match smulp randk _g with
  | None -> raise (Failure "bad r value, gives zero point after multiplication with generator")
  | Some(x,_) ->
      let r = mod_big_int x _n in
      if r = zero_big_int then
	raise (Failure "r value cannot be zero in signature");
      let s = mod_big_int (mult_big_int
			     (inv2 randk _n)
			     (add_big_int msgnum (mult_big_int r privk))) _n
      in
      if s = zero_big_int then raise (Failure "s value cannot be zero in signature");
      (r,s)

let check_sgn pubkey (r,s) msgnum =
  match pubkey with
  | None ->
      raise (Failure "zero point cannot be pubkey for signature")
  | Some(x,y) ->
      if lt_big_int r _n && lt_big_int s _n && lt_big_int msgnum _p then
	let sinv = inv2 s _n in
	let u1 = mod_big_int (mult_big_int msgnum sinv) _n in
	let u2 = mod_big_int (mult_big_int r sinv) _n in
	match addp (smulp u1 _g) (smulp u2 (Some(x,y))) with
	| None -> false
	| Some(compr,_) -> eq_big_int (mod_big_int compr _n) r
      else
	false
